(* L3 Compiler
 * IR Trees
 * Author: Kaustuv Chaudhuri <kaustuv+@cs.cmu.edu>
 * Modified: Alex Vaynberg <alv@andrew.cmu.edu>
 * Modified: Frank Pfenning <fp@cs.cmu.edu>
 * Modified: Joshua Wise <jwise>
 * Modified: Chris Lu <czl>
 *)

signature TREE =
sig
  datatype binop = ADD | SUB | MUL | DIV | MOD | LSH | RSH | LOGOR | LOGAND | BITOR | BITAND | BITXOR | NEQ | EQ | LT | GT | LE | GE | BE
  datatype unop = NEG | BITNOT | BANG

  type Blarg = int

  datatype exp = 
      CONST of Word32.word
    | TEMP of Temp.temp
    | ARG of Blarg (* I am j4cbo *)
    | BINOP of binop * exp * exp
    | UNOP of unop * exp
    | CALL of Ast.ident * exp list
    | MEMORY of exp
    | ALLOC of exp
    | COND of exp * exp * exp
    | STMVAR of stm list * exp
    | NULLPTR
  and stm =
      MOVE of exp * exp
    | RETURN of exp
    | EFFECT of exp
    | LABEL of Label.label
    | JUMPIFN of exp * Label.label
    | JUMP of Label.label
  and func =
      FUNCTION of Ast.ident * stm list

  type program = func list
end

structure Tree :> TREE =
struct
  datatype binop = ADD | SUB | MUL | DIV | MOD | LSH | RSH | LOGOR | LOGAND | BITOR | BITAND | BITXOR | NEQ | EQ | LT | GT | LE | GE | BE
  datatype unop = NEG | BITNOT | BANG

  type Blarg = int

  datatype exp = 
      CONST of Word32.word
    | TEMP of Temp.temp
    | ARG of Blarg (* I am j4cbo *)
    | BINOP of binop * exp * exp
    | UNOP of unop * exp
    | CALL of Ast.ident * exp list
    | MEMORY of exp
    | ALLOC of exp
    | COND of exp * exp * exp
    | STMVAR of stm list * exp
    | NULLPTR
  and stm =
      MOVE of exp * exp
    | RETURN of exp
    | EFFECT of exp
    | LABEL of Label.label
    | JUMPIFN of exp * Label.label
    | JUMP of Label.label
  and func =
      FUNCTION of Ast.ident * stm list

  type program = func list
end
